/*
 *****  lssig_mex.c written by Mark C Spaeth *****

 * A mex-able version of lssig which operates on files rather than 
 * structures, which works with lssig.m wrapper

 * This software comes with no warranty or support, and is distributed
 * under the GNU General Public License as described below.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the MxFree
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <float.h>
#include "mat.h"
#include "mex.h"

size_t (*fread_dummy)(void *data, size_t size, size_t n, FILE *fp);
size_t fread_we(void *data, size_t size, size_t n, FILE *fp);

unsigned int readheader(FILE *fp);
unsigned int blkhead[4];


void mexFunction(int nlhs, mxArray **plhs, int nrhs, const mxArray **prhs) {
  FILE *fp;
  unsigned int i,j,k,blk,blklen,seq;
  unsigned int nvars[4],ibuf[4],swlen;
  char c, *header, *hptr, *hptr2, filename[128];
  long nhead;
  int wrongway=0,hblks=0,inhead,nsigs,mlen;
  mxArray *mxtmp;
  char **sigs;

  if (nrhs!=1)
    mexErrMsgTxt("Error: Missing filename\nUsage: [sigs]=lssig_max('filename')");
  if (!mxIsChar(prhs[0]))
    mexErrMsgTxt("Error: Filename must be a string");
  mxGetString(prhs[0],filename,128);

  /* fprintf(stderr,"Filename: %s\n",filename); */

  fread_dummy=fread;
  if ((fp=fopen(filename,"rb"))==NULL)
    mexErrMsgTxt("Error: Cannot open file");

  /* Check endianness */
  if (fread(ibuf,4,1,fp)!=1)
    mexErrMsgTxt("Unexpected EOF in file");
  
  if (ibuf[0]==0x40000000) {
    fread_dummy=fread_we;
    wrongway=1;
  }
  rewind(fp);

  /* Validate block structure of file */
  hblks=0;
  nhead=0;
  inhead=1;
  while(inhead&&(blklen=readheader(fp))) {
    fseek(fp,blklen-8,SEEK_CUR);
    
    if (fread_dummy(ibuf,4,3,fp)!=3)
      mexErrMsgTxt("Unexpected EOF in file");
    else if (ibuf[2]!=blklen)
      mexErrMsgTxt("Bad block trailer in file");
    else {
      hblks++;
      nhead+=blklen;
    }
    if (ibuf[0]==0x23252624) inhead=0;
  }
  
  if (inhead)
    mexErrMsgTxt("Hit EOF before end of header");

  /*
    mexPrintf("%8d header blocks\n%8d header bytes\n",hblks,nhead);
  */

  /*
    If we made it here, the file structure is good
    No need to check for EOFs on fseek / fread anymore!
  */
    
  /* Read headers */
  rewind(fp);
  /* fprintf(stderr,"Header: %d\n",nhead+1); */
  if ((header=(char *)mxCalloc(nhead+1,sizeof(char)))==NULL) 
    mexErrMsgTxt("mxCalloc failed creating header");
  hptr=header;

  /* Parse headers */
  for (blk=0;blk<hblks;blk++) {
    blklen=readheader(fp);
    fread(hptr,1,blklen,fp);
    hptr+=blklen;
    fseek(fp,4,SEEK_CUR);      /* Pass block trailer */
  }

  /* fprintf(stderr,"Header read\n"); */

  /* Check post format */
  if (strncmp(header+0x10,"9601",4))
    mexErrMsgTxt("Error: file is not in post 9601 format");

  if ((j=sscanf(header,"%4d%4d%4d%4d",&nvars[0],&nvars[1],&nvars[2],&nvars[3]))!=4)
    mexErrMsgTxt("Error reading signal counts");

  /*
    mexPrintf("%4d independent variables\n",nvars[0]);
    mexPrintf("%4d dependent variables\n",nvars[1]);
    mexPrintf("%4d sweeps\n",nvars[2]);
  */
  
  /*
    Ignore sweep lengths  at offset 0x00b8 for now
    Ignore variable types at offset 0x0108 for this program (other than to count past them)
    Variable names start at 0x0108 + 0x08 * (total # vars & sweeps)
  */

  nsigs=nvars[0]+nvars[1]+nvars[2];
  sigs=(char **)mxCalloc(nsigs,sizeof(char *));

  hptr2=header+0x0108+0x08*(nvars[0]+nvars[1]);	   
  for (j=0;j<nsigs;j++) {
    if ((hptr=strtok(hptr2," \t\n"))==NULL)
      mexErrMsgTxt("Signal name not missing");
    else {
      for (k=0;k<strlen(hptr);k++) {
	c=hptr[k];
	if ((c<'0') || ((c>'9') && (c<'<')) || (c=='=') || ((c>'>') && (c<'A')) || ((c>'Z') && (c<'a')) || (c>'z'))
	  if (k<strlen(hptr)-1)
	    hptr[k]='_';
	  else
	    hptr[k]='\0';
	
      }
      /* Copy header into data structure */
      /*
	mexPrintf("%d : %s \n",strlen(hptr),hptr);
      */
      if((sigs[j]=(char *)mxCalloc(strlen(hptr)+1,sizeof(char)))==NULL)
	mexErrMsgTxt("mxCalloc Failed");
      strcpy(sigs[j],hptr);
      /*
	mexPrintf("Signal: %s\n",sigs[j]); 
      */
      hptr2=NULL;
    }
  }

  /* fprintf(stderr,"Setting plhs[0]\n"); */
  plhs[0]=mxCreateCharMatrixFromStrings(nsigs,(const char **)sigs);
  /* fprintf(stderr,"Set plhs[0]\n"); */


  for (j=0;j<nsigs;j++)
    mxFree(sigs[j]);
  mxFree(sigs);
  mxFree(header);
}


  unsigned int readheader(FILE *fp) {

  if (fread_dummy(blkhead, 4, 4, fp)!=4)
    return 0;
  
  if ((blkhead[0]!=0x04)||(blkhead[2]!=0x04))
    mexErrMsgTxt("Unexpected values in block header");
  
  return blkhead[3];
}

/* Wrong endian read (assume a 4-byte size) */
size_t fread_we(void *data, size_t size, size_t n, FILE *fp) {
  unsigned char *cptr,c;
  size_t i;

  if (fread(data,size,n,fp)!=n) return 0;

  for (i=0;i<=n;i++) {
    cptr=(unsigned char *)data+(4*i);
    c=cptr[0];
    cptr[0]=cptr[3];
    cptr[3]=c;
    c=cptr[1];
    cptr[1]=cptr[2];
    cptr[2]=c;
  }
  return(n);
}
